@@ -13,6 +13,7 @@
 #include <termios.h>
 #include <pwd.h>
 #include <sys/ioctl.h>
+#include <sys/select.h>
 
 #define PUTTY_DO_GLOBALS	       /* actually _define_ globals */
 #include "putty.h"
@@ -73,8 +74,6 @@ static Config cfg;
  */
 char *platform_default_s(const char *name)
 {
-    if (!strcmp(name, "X11Display"))
-	return dupstr(getenv("DISPLAY"));
     if (!strcmp(name, "TermType"))
 	return dupstr(getenv("TERM"));
     if (!strcmp(name, "UserName"))
@@ -130,10 +129,15 @@ void ldisc_update(void *frontend, int echo, int edit)
     else
 	mode.c_lflag &= ~ECHO;
 
-    if (edit)
+    if (edit) {
+	mode.c_iflag |= ICRNL;
 	mode.c_lflag |= ISIG | ICANON;
-    else
+    } else {
+	mode.c_iflag &= ~ICRNL;
 	mode.c_lflag &= ~(ISIG | ICANON);
+	mode.c_cc[VMIN] = 1;
+	mode.c_cc[VTIME] = 0;
+    }
 
     tcsetattr(0, TCSANOW, &mode);
 }
@@ -208,13 +212,14 @@ static void usage(void)
     printf("Usage: plink [options] [user@]host [command]\n");
     printf("       (\"host\" can also be a PuTTY saved session name)\n");
     printf("Options:\n");
+    printf("  -V        print version information and exit\n");
+    printf("  -pgpfp    print PGP key fingerprints and exit\n");
     printf("  -v        show verbose messages\n");
     printf("  -load sessname  Load settings from saved session\n");
     printf("  -ssh -telnet -rlogin -raw\n");
-    printf("            force use of a particular protocol (default SSH)\n");
+    printf("            force use of a particular protocol\n");
     printf("  -P port   connect to specified port\n");
     printf("  -l user   connect with specified username\n");
-    printf("  -m file   read remote command(s) from file\n");
     printf("  -batch    disable all interactive prompts\n");
     printf("The following options only apply to SSH connections:\n");
     printf("  -pw passw login with specified password\n");
@@ -228,9 +233,18 @@ static void usage(void)
     printf("  -A -a     enable / disable agent forwarding\n");
     printf("  -t -T     enable / disable pty allocation\n");
     printf("  -1 -2     force use of particular protocol version\n");
+    printf("  -4 -6     force use of IPv4 or IPv6\n");
     printf("  -C        enable compression\n");
     printf("  -i key    private key file for authentication\n");
+    printf("  -m file   read remote command(s) from file\n");
     printf("  -s        remote command is an SSH subsystem (SSH-2 only)\n");
+    printf("  -N        don't start a shell/command (SSH-2 only)\n");
+    exit(1);
+}
+
+static void version(void)
+{
+    printf("plink: %s\n", ver);
     exit(1);
 }
 
@@ -246,6 +260,7 @@ int main(int argc, char **argv)
     int errors;
     int use_subsystem = 0;
     void *ldisc, *logctx;
+    long now;
 
     ssh_get_line = console_get_line;
 
@@ -263,6 +278,7 @@ int main(int argc, char **argv)
      * Process the command line.
      */
     do_defaults(NULL, &cfg);
+    loaded_session = FALSE;
     default_protocol = cfg.protocol;
     default_port = cfg.port;
     errors = 0;
@@ -301,6 +317,11 @@ int main(int argc, char **argv)
 	    } else if (!strcmp(p, "-s")) {
                 /* Save status to write to cfg later. */
 		use_subsystem = 1;
+	    } else if (!strcmp(p, "-V")) {
+                version();
+            } else if (!strcmp(p, "-pgpfp")) {
+                pgp_fingerprints();
+                exit(1);
 	    } else if (!strcmp(p, "-o")) {
                 if (argc <= 1) {
                     fprintf(stderr,
@@ -345,7 +366,7 @@ int main(int argc, char **argv)
 		    strncpy(cfg.host, q, sizeof(cfg.host) - 1);
 		    cfg.host[sizeof(cfg.host) - 1] = '\0';
 		} else {
-		    char *r;
+		    char *r, *user, *host;
 		    /*
 		     * Before we process the [user@]host string, we
 		     * first check for the presence of a protocol
@@ -369,41 +390,46 @@ int main(int argc, char **argv)
 		    }
 
 		    /*
-		     * Three cases. Either (a) there's a nonzero
-		     * length string followed by an @, in which
-		     * case that's user and the remainder is host.
-		     * Or (b) there's only one string, not counting
-		     * a potential initial @, and it exists in the
-		     * saved-sessions database. Or (c) only one
-		     * string and it _doesn't_ exist in the
-		     * database.
+		     * A nonzero length string followed by an @ is treated
+		     * as a username. (We discount an _initial_ @.) The
+		     * rest of the string (or the whole string if no @)
+		     * is treated as a session name and/or hostname.
 		     */
 		    r = strrchr(p, '@');
 		    if (r == p)
 			p++, r = NULL; /* discount initial @ */
-		    if (r == NULL) {
-			/*
-			 * One string.
-			 */
+		    if (r) {
+			*r++ = '\0';
+			user = p, host = r;
+		    } else {
+			user = NULL, host = p;
+		    }
+
+		    /*
+		     * Now attempt to load a saved session with the
+		     * same name as the hostname.
+		     */
+		    {
 			Config cfg2;
-			do_defaults(p, &cfg2);
-			if (cfg2.host[0] == '\0') {
+			do_defaults(host, &cfg2);
+			if (loaded_session || cfg2.host[0] == '\0') {
 			    /* No settings for this host; use defaults */
-			    strncpy(cfg.host, p, sizeof(cfg.host) - 1);
+			    /* (or session was already loaded with -load) */
+			    strncpy(cfg.host, host, sizeof(cfg.host) - 1);
 			    cfg.host[sizeof(cfg.host) - 1] = '\0';
 			    cfg.port = default_port;
 			} else {
 			    cfg = cfg2;
-			    cfg.remote_cmd_ptr = cfg.remote_cmd;
 			}
-		    } else {
-			*r++ = '\0';
-			strncpy(cfg.username, p, sizeof(cfg.username) - 1);
-			cfg.username[sizeof(cfg.username) - 1] = '\0';
-			strncpy(cfg.host, r, sizeof(cfg.host) - 1);
-			cfg.host[sizeof(cfg.host) - 1] = '\0';
-			cfg.port = default_port;
 		    }
+
+		    if (user) {
+			/* Patch in specified username. */
+			strncpy(cfg.username, user,
+				sizeof(cfg.username) - 1);
+			cfg.username[sizeof(cfg.username) - 1] = '\0';
+		    }
+
 		}
 	    } else {
 		char *command;
@@ -454,7 +480,7 @@ int main(int argc, char **argv)
 
     /* See if host is of the form user@host */
     if (cfg.host[0] != '\0') {
-	char *atsign = strchr(cfg.host, '@');
+	char *atsign = strrchr(cfg.host, '@');
 	/* Make sure we're not overflowing the user field */
 	if (atsign) {
 	    if (atsign - cfg.host < sizeof cfg.username) {
@@ -496,7 +522,7 @@ int main(int argc, char **argv)
 	cfg.host[p1] = '\0';
     }
 
-    if (!*cfg.remote_cmd_ptr)
+    if (!cfg.remote_cmd_ptr && !*cfg.remote_cmd)
 	flags |= FLAG_INTERACTIVE;
 
     /*
@@ -548,7 +574,7 @@ int main(int argc, char **argv)
 	int nodelay = cfg.tcp_nodelay && isatty(0);
 
 	error = back->init(NULL, &backhandle, &cfg, cfg.host, cfg.port,
-			   &realhost, nodelay);
+			   &realhost, nodelay, cfg.tcp_keepalives);
 	if (error) {
 	    fprintf(stderr, "Unable to open connection:\n%s\n", error);
 	    return 1;
@@ -568,6 +594,7 @@ int main(int argc, char **argv)
     atexit(cleanup_termios);
     ldisc_update(NULL, 1, 1);
     sending = FALSE;
+    now = GETTICKCOUNT();
 
     while (1) {
 	fd_set rset, wset, xset;
@@ -628,7 +655,42 @@ int main(int argc, char **argv)
 	}
 
 	do {
-	    ret = select(maxfd, &rset, &wset, &xset, NULL);
+	    long next, ticks;
+	    struct timeval tv, *ptv;
+
+	    if (run_timers(now, &next)) {
+		ticks = next - GETTICKCOUNT();
+		if (ticks < 0) ticks = 0;   /* just in case */
+		tv.tv_sec = ticks / 1000;
+		tv.tv_usec = ticks % 1000 * 1000;
+		ptv = &tv;
+	    } else {
+		ptv = NULL;
+	    }
+	    ret = select(maxfd, &rset, &wset, &xset, ptv);
+	    if (ret == 0)
+		now = next;
+	    else {
+		long newnow = GETTICKCOUNT();
+		/*
+		 * Check to see whether the system clock has
+		 * changed massively during the select.
+		 */
+		if (newnow - now < 0 || newnow - now > next - now) {
+		    /*
+		     * If so, look at the elapsed time in the
+		     * select and use it to compute a new
+		     * tickcount_offset.
+		     */
+		    long othernow = now + tv.tv_sec * 1000 + tv.tv_usec / 1000;
+		    /* So we'd like GETTICKCOUNT to have returned othernow,
+		     * but instead it return newnow. Hence ... */
+		    tickcount_offset += othernow - newnow;
+		    now = othernow;
+		} else {
+		    now = newnow;
+		}
+	    }
 	} while (ret < 0 && errno == EINTR);
 
 	if (ret < 0) {
