@@ -98,6 +98,8 @@ static void sk_localproxy_close (Socket s)
     del234(localproxy_by_fromfd, ps);
     del234(localproxy_by_tofd, ps);
 
+    uxsel_del(ps->to_cmd);
+    uxsel_del(ps->from_cmd);
     close(ps->to_cmd);
     close(ps->from_cmd);
 
@@ -221,8 +223,8 @@ static int localproxy_select_result(int fd, int event)
 
 Socket platform_new_connection(SockAddr addr, char *hostname,
 			       int port, int privport,
-			       int oobinline, int nodelay, Plug plug,
-			       const Config *cfg)
+			       int oobinline, int nodelay, int keepalive,
+			       Plug plug, const Config *cfg)
 {
     char *cmd;
 
@@ -279,7 +281,7 @@ Socket platform_new_connection(SockAddr addr, char *hostname,
 	    close(i);
 	fcntl(0, F_SETFD, 0);
 	fcntl(1, F_SETFD, 0);
-	execl("/bin/sh", "sh", "-c", cmd, NULL);
+	execl("/bin/sh", "sh", "-c", cmd, (void *)NULL);
 	_exit(255);
     }
 
