@@ -12,6 +12,7 @@
 #include <utime.h>
 #include <errno.h>
 #include <assert.h>
+#include <glob.h>
 
 #include "putty.h"
 #include "psftp.h"
@@ -285,30 +286,71 @@ void close_directory(DirHandle *dir)
 
 int test_wildcard(char *name, int cmdline)
 {
-    /*
-     * On Unix, we currently don't support local wildcards at all.
-     * We will have to do so (FIXME) once PSFTP starts implementing
-     * mput, but until then we can assume `cmdline' is always set.
-     */
     struct stat statbuf;
 
-    assert(cmdline);
-    if (stat(name, &statbuf) < 0)
-	return WCTYPE_NONEXISTENT;
-    else
+    if (stat(name, &statbuf) == 0) {
 	return WCTYPE_FILENAME;
+    } else if (cmdline) {
+	/*
+	 * On Unix, we never need to parse wildcards coming from
+	 * the command line, because the shell will have expanded
+	 * them into a filename list already.
+	 */
+	return WCTYPE_NONEXISTENT;
+    } else {
+	glob_t globbed;
+	int ret = WCTYPE_NONEXISTENT;
+
+	if (glob(name, GLOB_ERR, NULL, &globbed) == 0) {
+	    if (globbed.gl_pathc > 0)
+		ret = WCTYPE_WILDCARD;
+	    globfree(&globbed);
+	}
+
+	return ret;
+    }
 }
 
 /*
- * Actually return matching file names for a local wildcard. FIXME:
- * we currently don't support this at all.
+ * Actually return matching file names for a local wildcard.
  */
 struct WildcardMatcher {
-    int x;
+    glob_t globbed;
+    int i;
 };
-WildcardMatcher *begin_wildcard_matching(char *name) { return NULL; }
-char *wildcard_get_filename(WildcardMatcher *dir) { return NULL; }
-void finish_wildcard_matching(WildcardMatcher *dir) {}
+WildcardMatcher *begin_wildcard_matching(char *name) {
+    WildcardMatcher *ret = snew(WildcardMatcher);
+
+    if (glob(name, 0, NULL, &ret->globbed) < 0) {
+	sfree(ret);
+	return NULL;
+    }
+
+    ret->i = 0;
+
+    return ret;
+}
+char *wildcard_get_filename(WildcardMatcher *dir) {
+    if (dir->i < dir->globbed.gl_pathc) {
+	return dupstr(dir->globbed.gl_pathv[dir->i++]);
+    } else
+	return NULL;
+}
+void finish_wildcard_matching(WildcardMatcher *dir) {
+    globfree(&dir->globbed);
+    sfree(dir);
+}
+
+int vet_filename(char *name)
+{
+    if (strchr(name, '/'))
+	return FALSE;
+
+    if (name[0] == '.' && (!name[1] || (name[1] == '.' && !name[2])))
+	return FALSE;
+
+    return TRUE;
+}
 
 int create_directory(char *name)
 {
@@ -321,55 +363,99 @@ char *dir_file_cat(char *dir, char *file)
 }
 
 /*
- * Wait for some network data and process it.
+ * Do a select() between all currently active network fds and
+ * optionally stdin.
  */
-int ssh_sftp_loop_iteration(void)
+static int ssh_sftp_do_select(int include_stdin, int no_fds_ok)
 {
     fd_set rset, wset, xset;
     int i, fdcount, fdsize, *fdlist;
     int fd, fdstate, rwx, ret, maxfd;
+    long now = GETTICKCOUNT();
 
     fdlist = NULL;
     fdcount = fdsize = 0;
 
-    /* Count the currently active fds. */
-    i = 0;
-    for (fd = first_fd(&fdstate, &rwx); fd >= 0;
-	 fd = next_fd(&fdstate, &rwx)) i++;
-
-    if (i < 1)
-	return -1;		       /* doom */
-
-    /* Expand the fdlist buffer if necessary. */
-    if (i > fdsize) {
-	fdsize = i + 16;
-	fdlist = sresize(fdlist, fdsize, int);
-    }
-
-    FD_ZERO(&rset);
-    FD_ZERO(&wset);
-    FD_ZERO(&xset);
-    maxfd = 0;
-
-    /*
-     * Add all currently open fds to the select sets, and store
-     * them in fdlist as well.
-     */
-    fdcount = 0;
-    for (fd = first_fd(&fdstate, &rwx); fd >= 0;
-	 fd = next_fd(&fdstate, &rwx)) {
-	fdlist[fdcount++] = fd;
-	if (rwx & 1)
-	    FD_SET_MAX(fd, maxfd, rset);
-	if (rwx & 2)
-	    FD_SET_MAX(fd, maxfd, wset);
-	if (rwx & 4)
-	    FD_SET_MAX(fd, maxfd, xset);
-    }
-
     do {
-	ret = select(maxfd, &rset, &wset, &xset, NULL);
-    } while (ret < 0 && errno == EINTR);
+
+	/* Count the currently active fds. */
+	i = 0;
+	for (fd = first_fd(&fdstate, &rwx); fd >= 0;
+	     fd = next_fd(&fdstate, &rwx)) i++;
+
+	if (i < 1 && !no_fds_ok)
+	    return -1;		       /* doom */
+
+	/* Expand the fdlist buffer if necessary. */
+	if (i > fdsize) {
+	    fdsize = i + 16;
+	    fdlist = sresize(fdlist, fdsize, int);
+	}
+
+	FD_ZERO(&rset);
+	FD_ZERO(&wset);
+	FD_ZERO(&xset);
+	maxfd = 0;
+
+	/*
+	 * Add all currently open fds to the select sets, and store
+	 * them in fdlist as well.
+	 */
+	fdcount = 0;
+	for (fd = first_fd(&fdstate, &rwx); fd >= 0;
+	     fd = next_fd(&fdstate, &rwx)) {
+	    fdlist[fdcount++] = fd;
+	    if (rwx & 1)
+		FD_SET_MAX(fd, maxfd, rset);
+	    if (rwx & 2)
+		FD_SET_MAX(fd, maxfd, wset);
+	    if (rwx & 4)
+		FD_SET_MAX(fd, maxfd, xset);
+	}
+
+	if (include_stdin)
+	    FD_SET_MAX(0, maxfd, rset);
+
+	do {
+	    long next, ticks;
+	    struct timeval tv, *ptv;
+
+	    if (run_timers(now, &next)) {
+		ticks = next - GETTICKCOUNT();
+		if (ticks <= 0)
+		    ticks = 1;	       /* just in case */
+		tv.tv_sec = ticks / 1000;
+		tv.tv_usec = ticks % 1000 * 1000;
+		ptv = &tv;
+	    } else {
+		ptv = NULL;
+	    }
+	    ret = select(maxfd, &rset, &wset, &xset, ptv);
+	    if (ret == 0)
+		now = next;
+	    else {
+		long newnow = GETTICKCOUNT();
+		/*
+		 * Check to see whether the system clock has
+		 * changed massively during the select.
+		 */
+		if (newnow - now < 0 || newnow - now > next - now) {
+		    /*
+		     * If so, look at the elapsed time in the
+		     * select and use it to compute a new
+		     * tickcount_offset.
+		     */
+		    long othernow = now + tv.tv_sec * 1000 + tv.tv_usec / 1000;
+		    /* So we'd like GETTICKCOUNT to have returned othernow,
+		     * but instead it return newnow. Hence ... */
+		    tickcount_offset += othernow - newnow;
+		    now = othernow;
+		} else {
+		    now = newnow;
+		}
+	    }
+	} while (ret < 0 && errno != EINTR);
+    } while (ret == 0);
 
     if (ret < 0) {
 	perror("select");
@@ -393,7 +479,58 @@ int ssh_sftp_loop_iteration(void)
 
     sfree(fdlist);
 
-    return 0;
+    return FD_ISSET(0, &rset) ? 1 : 0;
+}
+
+/*
+ * Wait for some network data and process it.
+ */
+int ssh_sftp_loop_iteration(void)
+{
+    return ssh_sftp_do_select(FALSE, FALSE);
+}
+
+/*
+ * Read a PSFTP command line from stdin.
+ */
+char *ssh_sftp_get_cmdline(char *prompt, int no_fds_ok)
+{
+    char *buf;
+    int buflen, bufsize, ret;
+
+    fputs(prompt, stdout);
+    fflush(stdout);
+
+    buf = NULL;
+    buflen = bufsize = 0;
+
+    while (1) {
+	ret = ssh_sftp_do_select(TRUE, no_fds_ok);
+	if (ret < 0) {
+	    printf("connection died\n");
+	    return NULL;	       /* woop woop */
+	}
+	if (ret > 0) {
+	    if (buflen >= bufsize) {
+		bufsize = buflen + 512;
+		buf = sresize(buf, bufsize, char);
+	    }
+	    ret = read(0, buf+buflen, 1);
+	    if (ret < 0) {
+		perror("read");
+		return NULL;
+	    }
+	    if (ret == 0) {
+		/* eof on stdin; no error, but no answer either */
+		return NULL;
+	    }
+
+	    if (buf[buflen++] == '\n') {
+		/* we have a full line */
+		return buf;
+	    }
+	}
+    }
 }
 
 /*
